/*
 * Copyright (c) 2017-2018, Texas Instruments Incorporated
 * All rights reserved.
 *
 * IMPORTANT: Your use of this Software is limited to those specific rights
 * granted under the terms of a software license agreement between the user
 * who downloaded the software, his/her employer (which must be your employer)
 * and Texas Instruments Incorporated (the "License"). You may not use this
 * Software unless you agree to abide by the terms of the License. The License
 * limits your use, and you acknowledge, that the Software may not be modified,
 * copied or distributed unless embedded on a Texas Instruments microcontroller
 * or used solely and exclusively in conjunction with a Texas Instruments radio
 * frequency transceiver, which is integrated into your product. Other than for
 * the foregoing purpose, you may not use, reproduce, copy, prepare derivative
 * works of, modify, distribute, perform, display or sell this Software and/or
 * its documentation for any purpose.
 *
 * YOU FURTHER ACKNOWLEDGE AND AGREE THAT THE SOFTWARE AND DOCUMENTATION ARE
 * PROVIDED "AS IS" WITHOUT WARRANTY OF ANY KIND, EITHER EXPRESS OR IMPLIED,
 * INCLUDING WITHOUT LIMITATION, ANY WARRANTY OF MERCHANTABILITY, TITLE,
 * NON-INFRINGEMENT AND FITNESS FOR A PARTICULAR PURPOSE. IN NO EVENT SHALL
 * TEXAS INSTRUMENTS OR ITS LICENSORS BE LIABLE OR OBLIGATED UNDER CONTRACT,
 * NEGLIGENCE, STRICT LIABILITY, CONTRIBUTION, BREACH OF WARRANTY, OR OTHER
 * LEGAL EQUITABLE THEORY ANY DIRECT OR INDIRECT DAMAGES OR EXPENSES
 * INCLUDING BUT NOT LIMITED TO ANY INCIDENTAL, SPECIAL, INDIRECT, PUNITIVE
 * OR CONSEQUENTIAL DAMAGES, LOST PROFITS OR LOST DATA, COST OF PROCUREMENT
 * OF SUBSTITUTE GOODS, TECHNOLOGY, SERVICES, OR ANY CLAIMS BY THIRD PARTIES
 * (INCLUDING BUT NOT LIMITED TO ANY DEFENSE THEREOF), OR OTHER SIMILAR COSTS.
 *
 * Should you have any questions regarding your right to use this Software,
 * contact Texas Instruments Incorporated at www.TI.com.
 *
 */

#include "tof_security.h"

#include "rtls_ctrl.h"
#include <ti/drivers/cryptoutils/cryptokey/CryptoKeyPlaintext.h>
#include <ti/drivers/TRNG.h>

#include <stdlib.h>
#include <stdbool.h>
#include <string.h>
#include <ti/sysbios/hal/Hwi.h>
#include <ti/sysbios/knl/Swi.h>

/*********************************************************************
 * LOCAL FUNCTIONS
 */
void TOFSecurity_freeBuffers(tofSecHandle_t *tofSecHandle);

/*******************************************************************************
 * @fn          TOFSecurity_paramsInit API
 *
 * @brief       This function to initialize the tof security parameters.
 *
 * input parameters
 *
 * @param       tofSecCfgPrms  - A pointer to the user-defined tof security preset values.
 *
 * @param       tofSecHandle  - A pointer to an empty tof security handle.
 *
 * output parameters
 *
 * @param       tofSecHandle  - A pointer to the tof security state.
 *
 * @return      int - NULL if failed, other if succeeded
 */
int TOFSecurity_paramsInit(tofSecHandle_t *tofSecHandle, tofSecCfgPrms_t *tofSecCfgPrms)
{
  uint16_t            sizeOfBuffToAlloc;

  // Sanity check
  if (tofSecHandle == NULL || tofSecCfgPrms == NULL)
  {
    return TOF_SEC_INVALID_PRMS_FAIL_E;
  }

  //  Verify params valid
  if ((0 == tofSecCfgPrms->syncWordSize) || (0 == tofSecCfgPrms->totalNumOfSyncWords))
  {
    return TOF_SEC_INVALID_PRMS_FAIL_E;
  }
  
  // Copy configuration parameters to handle
  memcpy(&(tofSecHandle->tofSecCfgParams),tofSecCfgPrms, sizeof(tofSecCfgPrms_t));

  // Init buffer/s sizes
  if (TOF_MODE_DBL_BUF == tofSecHandle->tofSecCfgParams.bUseDoubleBuffer)
  {
    tofSecHandle->numOfSyncWordsPerBuffer = TOF_SEC_DBL_BUFF_SIZE;
  }
  else
  {
    tofSecHandle->numOfSyncWordsPerBuffer = tofSecCfgPrms->totalNumOfSyncWords;
  }

  // Calculate size of buffer/s
  sizeOfBuffToAlloc = sizeof(uint8_t) * tofSecCfgPrms->syncWordSize * tofSecHandle->numOfSyncWordsPerBuffer;

  // Allocate first buffer
  RTLSUTIL_MALLOC(tofSecHandle->pSyncWordBuffer1, sizeOfBuffToAlloc);
  if (NULL == (tofSecHandle->pSyncWordBuffer1))
  {
    return TOF_SEC_ALLOC_FAIL_E;
  }

  // Allocate 2nd buffer?
  if (TOF_MODE_DBL_BUF == tofSecCfgPrms->bUseDoubleBuffer)
  {  
    RTLSUTIL_MALLOC(tofSecHandle->pSyncWordBuffer2, sizeOfBuffToAlloc);
    if (NULL == (tofSecHandle->pSyncWordBuffer2))
    {
      RTLSUTIL_FREE(tofSecHandle->pSyncWordBuffer1);
      return TOF_SEC_ALLOC_FAIL_E;
    }
  }

  // Allocate temp buffer
  RTLSUTIL_MALLOC(tofSecHandle->pTempSyncWordBuffer, sizeOfBuffToAlloc);
  if (NULL == (tofSecHandle->pTempSyncWordBuffer))
  {
    RTLSUTIL_FREE(tofSecHandle->pSyncWordBuffer2);
    RTLSUTIL_FREE(tofSecHandle->pSyncWordBuffer1);

    return TOF_SEC_ALLOC_FAIL_E;
  }

  return TOF_SEC_SUCCESS_E;
}

/*******************************************************************************
 * @fn          TOFSecurity_open API
 *
 * @brief       This function is used to open AESCTRDRBG driver and, when needed,
 *              to generate 128-bit random seed.
 *
 * input parameters
 *
 * @param       tofSecHandle  - A pointer to an empty tof security handle.
 *
 * @param       seed - NULL if need to generate seed, otherwise, a pointer to 128-bit random seed.
 *
 * output parameters
 *
 * @param       tofSecHandle  - A pointer to the tof security state.
 *
 * @return      int - NULL if failed, other if succeeded.
 */
int TOFSecurity_open(tofSecHandle_t *tofSecHandle, uint8_t *seed)
{
  AESCTRDRBG_Params   drbgParams;
  int_fast16_t        result;

  // If ToF master, generate seed
  if (seed == NULL)
  {
    TRNG_Handle         trngHandle;
    TRNG_Params         trngParams;
    CryptoKey           seedKey;

    // Init TRNG driver
    TRNG_init();

    TRNG_Params_init(&trngParams);
    trngParams.returnBehavior = TRNG_RETURN_BEHAVIOR_POLLING;

    // Open TRNG driver to get the handle
    trngHandle = TRNG_open(1, &trngParams);
    if (trngHandle == NULL)
    {
      TOFSecurity_freeBuffers(tofSecHandle);
      return TOF_SEC_GEN_FAIL_E;
    }

    // Init CryptoKey
    CryptoKeyPlaintext_initKey(&seedKey, tofSecHandle->seed, AESCTRDRBG_SEED_LENGTH_AES_128);

    // Generate seed
    result = TRNG_generateEntropy(trngHandle, &seedKey);
    if (result != TRNG_STATUS_SUCCESS)
    {
      TOFSecurity_freeBuffers(tofSecHandle);
      TRNG_close(trngHandle);
      return TOF_SEC_GEN_FAIL_E;
    }

    // Close TRNG driver
    TRNG_close(trngHandle);
  }
  else
  {
    memcpy(tofSecHandle->seed, seed, AESCTRDRBG_SEED_LENGTH_AES_128);
  }

  // Init DRBG driver
  AESCTRDRBG_init();

  AESCTRDRBG_Params_init(&drbgParams);
  drbgParams.keyLength = AESCTRDRBG_AES_KEY_LENGTH_128;
  drbgParams.reseedInterval = 0xFFFFFFFF;
  drbgParams.seed = tofSecHandle->seed;

  // Open DRBG driver
  tofSecHandle->drbgHandle = AESCTRDRBG_open(1, &drbgParams);

  if (tofSecHandle->drbgHandle == NULL)
  {
    TOFSecurity_freeBuffers(tofSecHandle);
    return TOF_SEC_GEN_FAIL_E;
  }

  return TOF_SEC_SUCCESS_E;
}

/*******************************************************************************
 * @fn          TOFSecurity_close API
 *
 * @brief       This function is used to close ToF Security and free buffers
 *
 * input parameters
 *
 * @param       tofSecHandle  - A pointer to an empty tof security handle.
 *
 * output parameters
 *
 * @return      int - NULL if failed, other if succeeded
 */
int TOFSecurity_close(tofSecHandle_t *tofSecHandle)
{
  // Free first buffer
  RTLSUTIL_FREE(tofSecHandle->pSyncWordBuffer1);

  // If double buffer is enabled, free the other one as well
  if (tofSecHandle->tofSecCfgParams.bUseDoubleBuffer == TOF_MODE_DBL_BUF)
  {
    RTLSUTIL_FREE(tofSecHandle->pSyncWordBuffer2);
  }

  // Close DRBG driver
  AESCTRDRBG_close(tofSecHandle->drbgHandle);

  return TOF_SEC_SUCCESS_E;
}

/*********************************************************************
* @fn       TOFSecurity_genSyncWords
*
* @brief    Will fill each Sync Word Buffer with <numOfSyncWordsPerBuffer> Sync Words
*
* input parameters
*
* @param   bufferToFill  - An enum describing which buffers to fill
*
* @param   tofSecHandle  - A pointer to the tof security state.
*
* output parameters
*
* @param   tofSecHandle  - A pointer to the tof security state.
*
* @return  NULL if failed, other if succeeded
*/
int TOFSecurity_genSyncWords(tofSecHandle_t *tofSecHandle, uint8_t bufferToFill)
{
  CryptoKey syncWordKey;
  int_fast16_t result;

  if ((TOF_SEC_FILL_BOTH_BUFFS == bufferToFill) || (TOF_SEC_FILL_1ST_BUFF == bufferToFill))
  {
    if (NULL != tofSecHandle->pSyncWordBuffer1)
    {
      // Initialise the CryptoKey
      CryptoKeyPlaintext_initKey(&syncWordKey,
                                 tofSecHandle->pSyncWordBuffer1,
                                 tofSecHandle->tofSecCfgParams.syncWordSize * tofSecHandle->numOfSyncWordsPerBuffer);
      // Generate syncwords
      result =  AESCTRDRBG_getBytes(tofSecHandle->drbgHandle, &syncWordKey);
      if (result != AESCTRDRBG_STATUS_SUCCESS)
      {
        return TOF_SEC_GEN_FAIL_E;
      }
    }
    else
    {
      return TOF_SEC_GEN_FAIL_E;
    }
  }
  
  if ((TOF_SEC_FILL_BOTH_BUFFS == bufferToFill) || (TOF_SEC_FILL_2ND_BUFF == bufferToFill))
  {
    if (tofSecHandle->tofSecCfgParams.bUseDoubleBuffer)
    {
      if (NULL != tofSecHandle->pSyncWordBuffer2)
      {
        // Initialise the CryptoKey
        CryptoKeyPlaintext_initKey(&syncWordKey,
                                   tofSecHandle->pSyncWordBuffer2,
                                   tofSecHandle->tofSecCfgParams.syncWordSize * tofSecHandle->numOfSyncWordsPerBuffer);
        // Generate syncwords
        result =  AESCTRDRBG_getBytes(tofSecHandle->drbgHandle, &syncWordKey);
        if (result != AESCTRDRBG_STATUS_SUCCESS)
        {
          return TOF_SEC_GEN_FAIL_E;
        }
      }
      else
      {
        return TOF_SEC_GEN_FAIL_E;
      }
    }
  }
  return TOF_SEC_SUCCESS_E;
}

/*********************************************************************
* @fn       TOFSecurity_freeBuffers
*
* @brief   This function free Sync Word Buffers
*
* @param   tofSecHandle  - A pointer to the tof security state.
*
* @return  none
*/
void TOFSecurity_freeBuffers(tofSecHandle_t *tofSecHandle)
{
  if (tofSecHandle->pSyncWordBuffer1 != NULL)
  {
    RTLSUTIL_FREE(tofSecHandle->pSyncWordBuffer1);
  }

  if (TOF_MODE_DBL_BUF == tofSecHandle->tofSecCfgParams.bUseDoubleBuffer &&
      tofSecHandle->pSyncWordBuffer2 != NULL)
  {
    RTLSUTIL_FREE(tofSecHandle->pSyncWordBuffer2);
  }

  if (tofSecHandle->pTempSyncWordBuffer != NULL)
  {
    RTLSUTIL_FREE(tofSecHandle->pTempSyncWordBuffer);
  }
}
